# -*- coding: utf-8 -*-

"""Console script for data_portal_explorer."""
import configparser
import json
import logging
import os
import random
import sys
import warnings
from concurrent import futures
from logging.config import fileConfig

import click
import pandas as pd
from ckanapi.errors import CKANAPIError
from pandas.io.json import json_normalize
from tqdm import tqdm

from data_portal_explorer.data_portal_explorer import (
    get_extensions,
    get_facets,
    get_number_of_packages,
    get_packages,
    get_resource,
)


@click.group(chain=True)
@click.argument("config", nargs=1, required=True, type=click.File("r"))
@click.argument(
    "dest", nargs=1, required=True, type=click.Path(file_okay=False, resolve_path=True)
)
@click.option(
    "--format",
    "fmt",
    default="json",
    show_default=True,
    type=click.Choice(["csv", "json"]),
)
@click.pass_context
def cli(ctx, config, dest, fmt):
    """Console script for data_portal_explorer."""
    click.echo("Data Portal Explorer")

    ctx.ensure_object(dict)

    try:
        parser = configparser.ConfigParser()
        parser.read_file(config)

        ctx.obj["PORTALS"] = get_portals(parser)
        ctx.obj["NAMESPACE"] = get_namespace(parser)
        ctx.obj["DATA_FORMATS"] = get_data_formats(parser)
        ctx.obj["WORKERS"] = get_workers(parser)
    except configparser.Error as e:
        click.secho(f"Failed to parse config file: {e.message}", fg="red")
        ctx.exit(code=-1)

    ctx.obj["DEST"] = dest
    ctx.obj["FORMAT"] = fmt

    try:
        os.makedirs(dest)
    except FileExistsError:
        pass

    fileConfig(parser.get("DEFAULT", "logging"))
    logger = logging.getLogger()
    logger.info("logger initialised")


def get_portals(config):
    active_portals = config.get("portals", "active").split()

    return [
        {"id": key, "themes": config.get(key, "themes"), "url": config.get(key, "url")}
        for key in active_portals
    ]


def get_namespace(config):
    return config.get(config.default_section, "namespace")


def get_data_formats(config):
    text_data_formats = config.get("data_formats", "text").split()
    excel_data_formats = config.get("data_formats", "excel").split()

    return {"text": text_data_formats, "excel": excel_data_formats}


def get_workers(config):
    try:
        return config.getint(config.default_section, "workers")
    except ValueError:
        return None


@cli.command()
@click.pass_context
def extensions(ctx):
    """Gets the available extensions."""
    click.echo("- Getting extensions")
    handle_command(ctx, "extensions", get_extensions)


def handle_command(ctx, name, func, *args):
    data = {}

    portals = ctx.obj["PORTALS"]

    with futures.ThreadPoolExecutor(max_workers=ctx.obj["WORKERS"]) as executor:
        future_to_portal = {
            executor.submit(func, portal, *args): portal for portal in portals
        }
        for future in tqdm(
            futures.as_completed(future_to_portal), total=len(future_to_portal.keys())
        ):
            portal = future_to_portal[future]
            portal_id = portal["id"]

            try:
                data[portal_id] = future.result()
            except CKANAPIError as e:
                click.secho(f" !error: get {name} for {portal_id}: {e}", fg="yellow")

    _save(ctx, name, data)


@cli.command()
@click.pass_context
def tags(ctx):
    """Gets the tags used by the datasets."""
    click.echo("- Getting tags")

    name = "tags"
    handle_command(ctx, name, get_facets, name)


@cli.command()
@click.pass_context
def themes(ctx):
    """Gets the themes used by the datasets."""
    click.echo("- Getting themes")

    name = "themes"
    handle_command(ctx, name, get_facets, name)


@cli.command()
@click.option("--rows", default=100, show_default=True, type=click.INT)
@click.option("--limit", default=0, show_default=True, type=click.INT)
@click.pass_context
def packages(ctx, rows, limit):
    """Gets packages."""
    click.echo("- Getting packages")

    portals = ctx.obj["PORTALS"]
    workers = ctx.obj["WORKERS"]
    ns = ctx.obj["NAMESPACE"]

    click.echo(" . preparing package requests")

    requests = get_packages_requests(portals, workers, ns, rows, limit)
    if not requests:
        click.secho(" ! failed to get packages requests")
        ctx.exit(code=-1)

    data = []

    click.echo(" . getting packages")

    logger = logging.getLogger()
    logger.info("packages")

    with futures.ThreadPoolExecutor(max_workers=workers) as executor:
        future_to_request = {
            executor.submit(get_packages, *request): request for request in requests
        }
        for future in tqdm(
            futures.as_completed(future_to_request), total=len(future_to_request.keys())
        ):
            request = future_to_request[future]

            try:
                data.extend(future.result())
            except (CKANAPIError, ConnectionError) as e:
                click.secho(f" ! error: get packages for {request}: {e}", fg="yellow")

    _save(ctx, "packages", data, normalise=True)


def get_packages_requests(portals, workers, namespace, rows, limit):
    requests = []

    with futures.ThreadPoolExecutor(max_workers=workers) as executor:
        future_to_portal = {
            executor.submit(get_number_of_packages, portal): portal
            for portal in portals
        }
        for future in tqdm(
            futures.as_completed(future_to_portal), total=len(future_to_portal.keys())
        ):
            portal = future_to_portal[future]
            portal_id = portal["id"]

            try:
                n = future.result()

                if limit > 0:
                    n = min(n, limit)

                start = 0
                number_of_requests = max(int(n / rows), 1)

                for i in range(number_of_requests):
                    requests.append([portal, namespace, start, rows])
                    start += rows
            except CKANAPIError as e:
                click.secho(
                    f" ! error: get number of packages for {portal_id}: {e}",
                    fg="yellow",
                )

    return requests


@cli.command()
@click.argument("packages_json", type=click.File("r"))
@click.pass_context
def resources(ctx, packages_json):
    """Extracts metadata from resources from previously downloaded
    packages metadata."""
    click.echo("- Extracting resources metadata")

    workers = ctx.obj["WORKERS"]
    namespace = ctx.obj["NAMESPACE"]
    data_formats = ctx.obj["DATA_FORMATS"]

    packages = json.load(packages_json)

    resources = []

    click.echo(" . preparing resource requests")

    with futures.ThreadPoolExecutor(max_workers=workers) as executor:
        future_to_package = {
            executor.submit(package.get, "resources"): package for package in packages
        }
        for future in tqdm(
            futures.as_completed(future_to_package), total=len(future_to_package.keys())
        ):
            package = future_to_package[future]
            result = future.result()

            for resource in result:
                resources.append([package, namespace, resource, data_formats])

    resources = random.sample(resources, k=len(resources))

    data = []

    click.echo(" . getting resources")

    logger = logging.getLogger()
    logger.info("resources")

    with futures.ThreadPoolExecutor(max_workers=workers) as executor:
        future_to_resource = {
            executor.submit(get_resource, *resource): resource for resource in resources
        }
        try:
            for future in tqdm(
                futures.as_completed(future_to_resource),
                total=len(future_to_resource.keys()),
            ):
                data.append(future.result())
        except KeyboardInterrupt:
            pass
        finally:
            _save(ctx, "resources", data, normalise=True)


def _save(ctx, filename, data, normalise=False):
    dst_path = os.path.join(ctx.obj["DEST"], filename)

    with open(f"{dst_path}.json", "w") as f:
        json.dump(data, f, indent=4, sort_keys=True)
        f.write("\n")
        f.close()

    if ctx.obj["FORMAT"] == "csv":
        df = json_normalize(data) if normalise else pd.read_json(f.name)
        df.index.name = filename
        df.to_csv(f"{dst_path}.csv")


if __name__ == "__main__":
    if not sys.warnoptions:
        warnings.simplefilter("ignore")

    sys.exit(cli())  # pragma: no cover



# -*- coding: utf-8 -*-

"""Main module."""

import http
import logging
import socket
import urllib
from collections import defaultdict
from functools import lru_cache

import pandas as pd
from ckanapi import RemoteCKAN
from ckanapi.errors import CKANAPIError

logger = logging.getLogger()


def get_extensions(portal):
    ckan = get_remote_ckan(portal["url"], get_only=True)

    r = ckan.action.status_show()
    if r:
        return {name: 1 for name in r["extensions"]}

    return {}


@lru_cache(maxsize=64)
def get_remote_ckan(portal_url, get_only=False):
    return RemoteCKAN(portal_url, get_only=True)


def get_facets(portal, name):
    facet_field = portal.get(name, name)
    ckan = get_remote_ckan(portal["url"], get_only=True)

    r = ckan.call_action(
        "package_search", {"facet.field": f'["{facet_field}"]', "facet.limit": -1}
    )

    if r and facet_field in r["facets"]:
        data = r["facets"][facet_field]
        return dict(sorted(data.items(), key=lambda t: t[0].lower()))

    return {}


def get_number_of_packages(portal):
    try:
        ckan = RemoteCKAN(portal["url"])
        r = ckan.action.package_search(rows=0)

        return r["count"]
    except CKANAPIError:
        return -1

    return 0


def get_packages(portal, namespace, start, rows):
    url = portal["url"]

    logger.info(f"get_packages: {url}; {start}-{rows}")

    ckan = get_remote_ckan(url)
    r = ckan.action.package_search(start=start, rows=rows)

    results = r.get("results")
    if not results:
        results = r.get("result")

    for package in results:
        package[f"{namespace}:portal"] = portal["id"]
        package[f"{namespace}:themes"] = package.get(portal["themes"])

    return results


def get_resources(package, namespace, data_formats):
    assert package is not None

    return [
        get_resource(package, namespace, resource, data_formats)
        for resource in package.get("resources")
    ]


def get_resource(package, namespace, resource, data_formats):
    resource[f"{namespace}:organisation"] = package.get("organization", "")
    resource[f"{namespace}:portal"] = package[f"{namespace}:portal"]
    resource[f"{namespace}:tags"] = get_package_tags(package)
    resource[f"{namespace}:themes"] = package.get(f"{namespace}:themes", "")

    is_open = package.get("isopen", False)
    data_format = resource.get("format").lower()
    url = resource.get("url")

    logger.info(
        f'get_resource: {resource["id"]}; {data_format}; ' f"is_open: {is_open}; {url}"
    )

    parsable_formats = data_formats["text"] + data_formats["excel"]

    if is_open and data_format in parsable_formats and not url.endswith(".zip"):
        resource.update(get_resource_data(url, data_format, data_formats, namespace))

    return resource


def get_package_tags(package):
    assert package is not None

    return ", ".join([t["display_name"] for t in package.get("tags")])


def get_resource_data(url, data_format, data_formats, namespace):
    assert url is not None

    data = defaultdict()

    try:
        if data_format in data_formats["excel"]:
            df = pd.read_excel(url)
        else:
            df = pd.read_csv(url)

        data[f"{namespace}:headers"] = get_headers(df)

        df = get_datetime_columns(df)
        data[f"{namespace}:max_date"] = str(get_max_date(df))
        data[f"{namespace}:min_date"] = str(get_min_date(df))
    except (
        ConnectionResetError,
        FileNotFoundError,
        UnicodeDecodeError,
        UnicodeEncodeError,
        ValueError,
        http.client.InvalidURL,
        pd.errors.EmptyDataError,
        pd.errors.ParserError,
        socket.gaierror,
        urllib.error.HTTPError,
        urllib.error.URLError,
    ) as e:
        data[f"{namespace}:error_message"] = str(e)
        data[f"{namespace}:error_url"] = url

    return data


def get_headers(df):
    assert df is not None

    return ", ".join(df.columns)


def get_datetime_columns(df):
    assert df is not None

    df = convert_columns_to_datetime(df)

    return df.select_dtypes(include=["datetime"])


def convert_columns_to_datetime(df):
    assert df is not None

    return df.apply(
        lambda col: pd.to_datetime(col, errors="ignore", infer_datetime_format=True)
        if col.dtypes == object
        else col,
        axis=0,
    )


def get_max_date(df):
    assert df is not None

    if len(df.columns) == 0:
        return None

    try:
        timestamp = df.max(axis=0).max()
    except TypeError:
        return None

    if isinstance(timestamp, pd.Timestamp):
        return timestamp.date()

    return None


def get_min_date(df):
    assert df is not None

    if len(df.columns) == 0:
        return None

    try:
        timestamp = df.min(axis=0).min()
    except TypeError:
        return None

    if isinstance(timestamp, pd.Timestamp):
        return timestamp.date()

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import os
import sys

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

with open("requirements.txt") as requirements_file:
    requirements = requirements_file.read()

setup_requirements = []

test_requirements = []

if sys.argv[-1] == "publish":
    try:
        import wheel

        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system("python setup.py sdist")
    os.system("python setup.py bdist_wheel")
    os.system("twine upload dist/*")
    sys.exit()


setup(
    author="Kings Digital Lab",
    author_email="jmvieira@pm.me",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    description="""This package contains tools to extract metadata from CKAN
    instances which can be used to study data portals as infrastructures from
    the perspective of social/cultural research.""",
    entry_points={
        "console_scripts": [
            "data_portal_explorer=data_portal_explorer.cli:cli",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="data_portal_explorer",
    name="data_portal_explorer",
    packages=find_packages(include=["data_portal_explorer"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/kingsdigitallab/DataPortalExplorer",
    version="0.1.9",
    zip_safe=False,
)
